/**
 * @file    shellmatta_transport.c
 * @brief   transport layer functions of shellmatta
 * @author  Simon Fischer <fischer.simon.1991@gmail.com>
 */

#include "shellmatta_transport.h"
#include "shellmatta_crc.h"
#include <string.h>

/* init global variables */
uint8_t protocolVersion                     = 0u;
shellmatta_transport_packet_t packetType    = PACKET_DATA;
uint32_t payloadLength                      = 0u;
uint8_t source                              = 0u;
uint8_t destination                         = 0u;
uint32_t crc32                              = 0u;
uint8_t payloadCounter                      = 0u;
uint8_t crcCounter                          = 0u;
uint8_t packetSequenceCounter_h2s           = 0u;
uint8_t packetSequenceCounter_s2h           = 0u;
shellmatta_transport_layer_t transportLayerInst = {
    0u,
    0u,
    STATE_GET_SOH,
    false,
    false,
    false,
    NULL
};
char payloadBuffer[SHELLMATTA_PAYLOAD_MAXLENGTH];

/**
 * @brief       Initializes the transportLayerInst
 * @return      errorcode   #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_init_transport_inst(void)
{
    transportLayerInst.h2s_sequenceCnt = 0u;
    transportLayerInst.s2h_sequenceCnt = 0u;
    transportLayerInst.state = STATE_GET_SOH;
    transportLayerInst.mandatory = false;
    transportLayerInst.active = false;
    transportLayerInst.continueStep = false;
    memset(payloadBuffer, 0u, SHELLMATTA_PAYLOAD_MAXLENGTH);

    return SHELLMATTA_OK;
}

/**
 * @brief       Resets all values of tranportLayerInst except for sequence counters
 * @return      errorcode   #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_reset_transport(void)
{
    transportLayerInst.state = STATE_GET_SOH;
    protocolVersion = 0u;
    packetType      = PACKET_DATA;
    payloadLength   = 0u;
    source          = 0u;
    destination     = 0u;
    crc32           = 0u;
    payloadCounter  = 0u;
    crcCounter      = 0u;
    transportLayerInst.active = false;
    memset(payloadBuffer, 0u, SHELLMATTA_PAYLOAD_MAXLENGTH);

    return SHELLMATTA_OK;
}

/**
 * @brief       processes the passed amount of data
 * @param[in]   data    pointer to input data to process
 * @return      errorcode   #SHELLMATTA_OK\n
 *                          #SHELLMATTA_ERROR   in case of crc error
 */
shellmatta_retCode_t shellmatta_handle_transport_fsm(char *data)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;

    switch (transportLayerInst.state)
    {
        /* look for start of header */
    case STATE_GET_SOH:
        /* if start of header is found, continue transport layer fsm */
        if (SHELLMATTA_START_OF_HEADER == *data)
        {
            transportLayerInst.state = STATE_GET_PROTOCOL_VERSION;
            transportLayerInst.active = true;
        }
        else if (0x00 != *data)
        {
            /* if SOH is not found, handle data as manual input */
            transportLayerInst.state = STATE_MANUAL_INPUT;
            transportLayerInst.active = false;
        }
        break;

    case STATE_MANUAL_INPUT:
        break;

    /* look for protocol version */
    case STATE_GET_PROTOCOL_VERSION:
        /* check protocol version */
        if (PROTOCOL_VERSION == *data)
        {
            transportLayerInst.state = STATE_GET_PACKET_TYPE;
        }
        else
        {
            /* reset fsm in case of wrong protocol version */
            transportLayerInst.state = STATE_GET_SOH;
            transportLayerInst.active = false;
        }
        break;

    /* read packet type */
    case STATE_GET_PACKET_TYPE:
        transportLayerInst.state = STATE_GET_PAYLOAD_LENGTH;
        break;

    /* look for length of payload */
    case STATE_GET_PAYLOAD_LENGTH:
        transportLayerInst.state = STATE_GET_SOURCE;
        break;

    /* look for source */
    case STATE_GET_SOURCE:
        transportLayerInst.state = STATE_GET_DESTINATION;
        break;

    /* look for destination */
    case STATE_GET_DESTINATION:
        transportLayerInst.state = STATE_GET_H2S_SEQUENCE_CNT;
        break;

    /* look for h2s sequence counter*/
    case STATE_GET_H2S_SEQUENCE_CNT:
        transportLayerInst.state = STATE_GET_S2H_SEQUENCE_CNT;
        break;

    /* look for s2h sequence counter */
    case STATE_GET_S2H_SEQUENCE_CNT:
        if (packetType == PACKET_SEQ_CNT_REQUEST)
        {
            transportLayerInst.state = STATE_GET_CRC;
        }
        else
        {
            transportLayerInst.state = STATE_GET_PAYLOAD;
        }
        break;

    /* read payload for previously read length of payload */
    case STATE_GET_PAYLOAD:
        if (    (payloadLength <= payloadCounter)
            &&  (true == transportLayerInst.active))
        {
            transportLayerInst.state = STATE_GET_CRC;
        }
        break;

    /* read crc32 for four bytes */
    case STATE_GET_CRC:
        /* if crc32 is read, start computing */
        if (SHELLMATTA_LENGTH_CRC <= crcCounter)
        {
            /* for crc computation only */
            char crcdata[SHELLMATTA_PAYLOAD_MAXLENGTH + SHELLMATTA_HEADER_LENGTH];
            memset(crcdata, 0u, SHELLMATTA_PAYLOAD_MAXLENGTH + SHELLMATTA_HEADER_LENGTH);
            crcdata[SHELLMATTA_POS_SOH]         = SHELLMATTA_START_OF_HEADER;
            crcdata[SHELLMATTA_POS_PROT_VER]    = protocolVersion;
            crcdata[SHELLMATTA_POS_PACKET_TYPE] = packetType;
            crcdata[SHELLMATTA_POS_PAYLOAD_LEN] = payloadLength;
            crcdata[SHELLMATTA_POS_SRC]         = source;
            crcdata[SHELLMATTA_POS_DST]         = destination;
            crcdata[SHELLMATTA_POS_CNT_H2S]     = packetSequenceCounter_h2s;
            crcdata[SHELLMATTA_POS_CNT_S2H]     = packetSequenceCounter_s2h;
            strncat(&crcdata[SHELLMATTA_POS_PAYLOAD], payloadBuffer, payloadLength);

            uint32_t refCrc;

            refCrc = crc32Calc(crcdata, SHELLMATTA_HEADER_LENGTH + payloadLength);

            /* if crc is correct, further handling of data depends on type of packet */
            if (crc32 == refCrc)
            {
                /* reset fsm by default */
                transportLayerInst.state = STATE_GET_SOH;

                /* crc is correct */
                transportLayerInst.h2s_sequenceCnt++;

                switch (packetType)
                {
                case PACKET_DATA:
                    transportLayerInst.state = STATE_PROCESS_PAYLOAD;
                    break;
                
                case PACKET_SEQ_CNT_REQUEST:
                    /* change packet type to response */
                    packetType = PACKET_SEQ_CNT_RESPOND;
                    /* send out packet with no payload */
                    shellmatta_write_transport("", 0u);
                    break;

                case PACKET_SEQ_CNT_RESPOND:
                    break;
                
                case PACKET_MAX_BUFFERSIZE_REQUEST:
                    break;

                case PACKET_MAX_BUFFERSIZE_RESPOND:
                    break;
                
                case PACKET_SEARCH_DEVICE_REQUEST:
                    break;
                
                case PACKET_SEARCH_DEVICE_RESPOND:
                    break;

                case PACKET_SET_ADDRESS_REQUEST:
                    break;

                case PACKET_SET_ADDRESS_RESPOND:
                    break;

                /* wrong packet type */
                default:
                    /* undo sequence counter increment */
                    transportLayerInst.h2s_sequenceCnt--;
                    break;
                }
            }
            else
            {
                transportLayerInst.state = STATE_GET_SOH;
                /* crc is incorrect */
                ret = SHELLMATTA_ERROR;
            }

            /* reset crc variables */
            crcCounter = 0u;
            crc32 = 0u;
        }
        break;

    /* start to process payload as if it was entered manually */
    case STATE_PROCESS_PAYLOAD:
        transportLayerInst.state = STATE_GET_SOH;
        protocolVersion = 0u;
        packetType      = PACKET_DATA;
        payloadLength   = 0u;
        source          = 0u;
        destination     = 0u;
        crc32           = 0u;
        payloadCounter  = 0u;
        crcCounter      = 0u;
        transportLayerInst.active = false;
        memset(payloadBuffer, 0u, SHELLMATTA_PAYLOAD_MAXLENGTH);
        break;
    
    default:
        break;
    }

    return ret;
}

/**
 * @brief       Wrapper function for the write-function of shellmatta handle
 * 
 * This function is used to transmit data with the tranport layer protocol.\n
 * The input data is copied into a new array along with the header and crc32.\n
 * The resulting buffer is the forwarded to the original write function.\n
 * 
 * @note        If length of data exceeds #UINT8_MAX, data is sent in multiple packets.\n
 * 
 * @param[in]   data    pointer to input data to process
 * @return      errorcode   #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_write_transport(const char* data, uint32_t length)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    
    /* create buffer for data + header + crc with maximum size */
    uint8_t outputBuffer[   SHELLMATTA_HEADER_LENGTH +      /* length of header */
                            SHELLMATTA_PAYLOAD_MAXLENGTH +  /* max length of payload */
                            SHELLMATTA_LENGTH_CRC];         /* length of crc */
    
    uint16_t outputBufferSize = sizeof(outputBuffer);
    uint32_t outPayloadLength = length;

    /* handle data with length bigger than max length */
    uint16_t processedLength = 0u;
    uint16_t splitLength = 0u;
    uint16_t splitPayloadLength = 0u;
    uint16_t restOfPayload = 0u;

    /* foot-controlled loop to send packets without payload length */
    do
    {
        /* compute length of next payload split */
        restOfPayload = (outPayloadLength - processedLength);
        if (restOfPayload > SHELLMATTA_PAYLOAD_MAXLENGTH)
        {
            splitLength = outputBufferSize;
        }
        else
        {
            splitLength = (restOfPayload + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_LENGTH_CRC) % outputBufferSize;
        }
        splitPayloadLength = splitLength - SHELLMATTA_HEADER_LENGTH - SHELLMATTA_LENGTH_CRC;

        /* fill buffer with header data */
        outputBuffer[SHELLMATTA_POS_SOH] = SHELLMATTA_START_OF_HEADER;           /* start of header */
        outputBuffer[SHELLMATTA_POS_PROT_VER] = PROTOCOL_VERSION;                     /* protocol version */
        outputBuffer[SHELLMATTA_POS_PACKET_TYPE] = packetType;                           /* packet type */
        outputBuffer[SHELLMATTA_POS_PAYLOAD_LEN] = splitPayloadLength;                   /* payload length */
        outputBuffer[SHELLMATTA_POS_SRC] = 0x00u;                                /* source */
        outputBuffer[SHELLMATTA_POS_DST] = 0x00u;                                /* destination */
        outputBuffer[SHELLMATTA_POS_CNT_H2S] = transportLayerInst.h2s_sequenceCnt;   /* sequence counter host to shellmatta */
        outputBuffer[SHELLMATTA_POS_CNT_S2H] = ++transportLayerInst.s2h_sequenceCnt; /* sequence counter shellmatta to host */

        /* skip copying of payload in case of sequence counter response */
        if (packetType != PACKET_SEQ_CNT_RESPOND)
        {
            memcpy(&outputBuffer[SHELLMATTA_POS_PAYLOAD], data + processedLength, splitPayloadLength);
        }

        uint32_t outCrc = crc32Calc((char*) outputBuffer, SHELLMATTA_HEADER_LENGTH + splitPayloadLength);

        /* append crc to end of payload */
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH] = (uint8_t)(outCrc);
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_SHIFT_BY_1] = (uint8_t)(outCrc >> SHELLMATTA_SHIFT_BY_8);
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_SHIFT_BY_2] = (uint8_t)(outCrc >> SHELLMATTA_SHIFT_BY_16);
        outputBuffer[splitPayloadLength + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_SHIFT_BY_3] = (uint8_t)(outCrc >> SHELLMATTA_SHIFT_BY_24);

        /* use original write function to send full buffer */
        ret = transportLayerInst.originalWrite((char*) outputBuffer, splitLength);
        processedLength += splitPayloadLength;
    }
    while (processedLength < outPayloadLength);
    return ret;
}
